__doc__   = """Photo object"""
__version = '$ Release 0.1.0 $'[-7:-2]

############################################################################### 
#
# Copyright (c) 1999 Andrew Lahser
# All rights reserved. Written by Andrew Lahser andrew@apl-software.com
# 
# Redistribution and use in source and binary forms, with or without 
# modification, are permitted provided that the following conditions 
# are met: 
# 1. Redistributions of source code must retain the above copyright 
#    notice, this list of conditions and the following disclaimer. 
# 2. Redistributions in binary form must reproduce the above copyright 
#    notice, this list of conditions and the following disclaimer in the 
#    documentation and/or other materials provided with the distribution. 
# 3. The name of the author may not be used to endorse or promote products 
#    derived from this software without specific prior written permission 
# 
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR 
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
# IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, 
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT 
# NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF 
# THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
#   
#  In accordance with the license provided for by the software upon 
#  which some of the source code has been derived or used, the following 
#  acknowledgement is hereby provided : 
# 
#      "This product includes software developed by Digital Creations 
#      for use in the Z Object Publishing Environment 
#      (http://www.zope.org/)." 
#  
############################################################################### 
   

from OFS.content_types import guess_content_type
from OFS.Image import File, Image, cookId, Pdata
from Globals import HTMLFile, MessageDialog
from OFS.PropertyManager import PropertyManager
from AccessControl.Role import RoleManager
from webdav.common import rfc1123_date
from OFS.SimpleItem import Item_w__name__
from cStringIO import StringIO
from Globals import Persistent
from Acquisition import Implicit
from DateTime import DateTime
from types import IntType, StringType, TupleType
from urllib import quote, unquote                                                        
import Globals, string, struct, OFS.content_types
import PIL

StringType=type('')

manage_addPhotoForm=HTMLFile('PhotoAdd',globals(),Kind='Photo',kind='Photo')

def manage_addPhoto(self, id, file, title='', displays=None, precondition='', 
                    content_type='', REQUEST=None):
    """
    Add a new Photo object.
    """
    id, title = cookId(id, title, file)

    self=self.this()

    # First, we create the photo without data:
    self._setObject(id, Photo(id,title,'',displays,content_type,precondition))
        
    # Now we "upload" the data.  By doing this in two steps, we
    # can use a database trick to make the upload more efficient.
    self._getOb(id).manage_upload(file)
    
    if REQUEST:
        try:    url=self.DestinationURL()
        except: url=REQUEST['URL1']
        REQUEST.RESPONSE.redirect('%s/manage_main' % url)
    return id

 
class Photo(Image):
    """
    Photo objects are images. Each Photo will save itself at a variety
    of display sizes for fast and easy retrieval from the web.
    """
    meta_type='Photo'
    _properties=({'id':'title', 'type': 'string'},
                 {'id':'content_type', 'type':'string'},
                 {'id':'height', 'type':'string'},
                 {'id':'width', 'type':'string'},
                 )
    
    manage_options=({'label':'Edit', 'action':'manage_main'},
                    {'label':'Upload', 'action':'manage_uploadForm'},
                    {'label':'Displays', 'action':'manage_displays'},
                    {'label':'Properties', 'action':'manage_propertiesForm'},
                    {'label':'View', 'action':'view_image_or_file'},
                    {'label':'Security', 'action':'manage_access'},
                    )

    manage_displays=HTMLFile('displays',globals())

    def __init__(self, id, title, file, 
                 displays = None, content_type = '', precondition = ''):
        self.__name__=id
        self.title=title
        self.precondition=precondition
       
        data, size = self._read_data(file)
        content_type=self._get_content_type(file, data, id, content_type)
	if displays == None:
            self.displays={ 
	      'thumbnail':{'actual' : (0,0), 'target' : (128,128),  'size' : 0}
             ,'preview'  :{'actual' : (0,0), 'target' : (370,263),  'size' : 0}
             ,'webtv'    :{'actual' : (0,0), 'target' : (544,378),  'size' : 0}
             ,'xsmall'   :{'actual' : (0,0), 'target' : (623,278),  'size' : 0}
             ,'small'    :{'actual' : (0,0), 'target' : (783,398),  'size' : 0}
             ,'medium'   :{'actual' : (0,0), 'target' : (1007,566), 'size' : 0}
             ,'large'    :{'actual' : (0,0), 'target' : (1135,668), 'size' : 0}
             ,'xlarge'   :{'actual' : (0,0), 'target' : (1263,822), 'size' : 0}
              }
	else:
            self.displays=displays
        self.update_data(data, content_type, size)
    
    def __call__(self, REQUEST=None, display=None, 
                 pdcookie=None, height=None, 
		 width=None, alt=None, **args):
        """
        Generate an HTML IMG tag for this photo, with customization. 
	The display size will be selected from 
	1. DTML parameter
	2. Users pdcookie
	3. DTML specified width and height (this is dynamicaly generated).
	4. Default to preview
	5. Display original image.
	Other tags may be specified as keyword=value pairs.
        """
	# Check for a display parmameter in DTML
	if display and self.displays.has_key(display):
            (width, height) = self.displays[display]['actual']
	    string ='<img src="%s?display=%s" width="%s" height="%s"' % (
	        self.absolute_url(), display, width, height
		)
	# Check the cookie for the preferred display size
	elif (REQUEST and REQUEST.cookies.has_key('pd') and
	      self.displays.has_key(REQUEST.cookies['pd'])):
            display=REQUEST.cookies['pd']
            (width, height) = self.displays[display]['actual']
	    string ='<img src="%s?display=%s" width="%s" height="%s"' % (
	        self.absolute_url(), display, width, height
		)
        # Check for dynamic size specification in DTML, 
	# this can be SLOW, because the image is PIL'd for every request.
	elif width or height:
	    (width, height) = self._preserveAspectRatio(width, height)
            string='<img src="%s/resize?width=%s&height=%s"' % (
	        self.absolute_url(), width, height
		)
	    string ='%s width="%s" height="%s"' % (
	        string, width, height
		)
	# Default to preview display if no cookie
        elif self.displays.has_key('preview'):
            (width, height) = self.displays['preview']['actual']
	    string ='<img src="%s" width="%s" height="%s"' % (
	        self.absolute_url(), width, height
		)
	# Finally, use original image as a last resort
	else:
	    string = '<img src="%s" width="%s" height="%s"' % (
	        self.absolute_url(), self.width, self.height
                )
        # Address remaining attributes
        if alt is None:
            alt=self.title_or_id()
        if alt:
            string = '%s alt="%s"' % (string, alt)
        for key in args.keys():
            value = args.get(key)
            string = '%s %s="%s"' % (string, key, value)
	string = string +'>'
        # Add Cookie Bar
	if pdcookie and REQUEST:
            pd='<table border="0" cellpadding="0" cellspacing="0"><tr>'
	    pd='%s<td align="center" colspan="%s">' % (pd, len(self.displays))
	    pd='%s%s</td></tr><tr>' % (pd, string)
	    for d in self.displayMap():
	        if d['id'] in ('thumbnail','preview'): continue
                pd='%s<td align="center"><font size="1"> ' % (pd)
	        if (REQUEST.cookies.has_key('pd') and 
		    d['id'] == REQUEST.cookies['pd']):
                    pd='%s %s ' % (pd, d['id'])
		else:
                    pd='%s <a href="%s/setcookie?display=%s&redir=%s">' % (
	                pd, self.absolute_url(), d['id'], quote(REQUEST.URL)
                        )
                    pd='%s%s</a> ' % (pd, d['id'])
                pd='%s </font></td>' % (pd)
            else:
	        pd=pd + '<td></td>'
	    pd='%s</tr></table>' % (pd)
	    string = pd
        return string 

    def displayMap(self):
        """Return a tuple of photo mappings"""
	displayMap = []
        for display in self.displays.keys():
	    (width, height) = self.displays[display]['target']
	    (actual_width, actual_height) = self.displays[display]['actual']
	    size = self.displays[display]['size']
            displayMap = displayMap +  [{  'id' : display
	                                 , 'width' : width
					 , 'height' : height
	                                 , 'actual_width' : actual_width
					 , 'actual_height' : actual_height
					 , 'size' : size}]
	displayMap.sort(lambda a,b: cmp(a['size'],b['size']))
	return tuple(displayMap)

    def setcookie(self, display=None, redir=None, REQUEST=None):
        """
	Set a cookie and redirect browser to the requesting page
	"""
	if REQUEST and display:
	    REQUEST.RESPONSE.setCookie('pd',display,path='/')
	if REQUEST and redir:
            REQUEST.RESPONSE.redirect(redir)

    tag=__call__

    def update_data(self, data, content_type = None, size = None):
        """
        Replaces the current contents of the Photo object with file.
        """
        Image.update_data(self, data, content_type, size)
        self._generateDisplays(REQUEST=None)


    def _addDisplay(self, display, width, height):
        """ Add a display for this photo """
        if self.displays.has_key(display):
            raise 'Bad Request', 'Invalid or duplicate display name'
	self.displays[display]= {
	    'actual' : (0,0),
	    'target' : (width,height),
	    'size' : 0}
	self._generateDisplay(display)
 
    def _editDisplay(self, display, width, height):
        """ Modify a display for this photo """
        self.displays[display]['target'] = (width, height)
	self._generateDisplay(display)

    def _delDisplay(self, display):
        """ Delete a display from this photo """
        delattr(self, 'data_' + display)
	displays = self.displays
	del displays[display]
	self.displays = displays

    def _generateDisplays(self, REQUEST=None):
        """ Recreate all of the display sizes for thi photo """
	if (self.size > 0):
            for display in self.displays.keys():
                (width, height) = self.displays[display]['target']
                self._generateDisplay(display)
        if REQUEST:
            REQUEST['RESPONSE'].redirect(self.absolute_url()+'/manage_displays')

    def _generateDisplay(self, display):
        """ Recreate one display for a Photo """
	if (self.size <= 0):
	    return None
	(width, height) = self.displays[display]['target']
        (width, height) = self._preserveAspectRatio(width, height)
	setattr(self, 'data_' + display, self.resize(width, height))
        self.displays[display]['actual'] = (width, height)
	self.displays[display]['size'] = len(getattr(self, 'data_' + display))


    def _preserveAspectRatio(self, width, height):
        """ 
	Return largest width and height while preserving the aspect ratio 
	"""
	original_width=int(self.width)
	original_height=int(self.height)
	if width: width=int(width)
	if height: height=int(height)
	if (self.size <= 0) or (height is None and width is None): 
            width = None
            height = None
        elif height is None or ( width and 
	      (height > original_height * width / original_width)):
            height = original_height * width / original_width
        else:
            width =  original_width * height / original_height
        return(width, height)

    def resize(self, width=None, height=None, REQUEST = None):
        """Return a scaled image"""
        if (self.size <= 0) or (height is None or width is None): 
            return None
        # Unwrap PData if size bigger than 1<<16
	if type(self.data) == StringType:
            im = PIL.Image.open(StringIO(self.data))
	else:
            r = [self.data.data]
            next = self.data.next
            while next is not None:
	        self=next
                r.append(self.data)
		next=self.next
            r = string.join(r,'')
            im = PIL.Image.open(StringIO(r))
	fmt = im.format
        im = im.resize((int(width),int(height)))
        outFile = StringIO()
        im.save(outFile, fmt)
        outFile.seek(0)
        outString = outFile.read()
	# Set Response Headers if HTTP request
        if REQUEST:
            REQUEST.RESPONSE.setHeader('Last-Modified', rfc1123_date(self._p_mtime))
            REQUEST.RESPONSE.setHeader('Content-Type', self.content_type)
        return outString

#   Web Interface
    def manage_editDisplay(self, REQUEST = None):
        """ Edit displays through the web"""
        for display in self.displays.keys():
	    width = REQUEST.get('width-' + display)
	    width = int(width)
	    height = REQUEST.get('height-' + display)
	    height = int(height)
            self._editDisplay(display, width, height)
        return MessageDialog(
               title  ='Success!',
               message='Your changes have been saved',
               action ='manage_displays')
	
    def manage_addDisplay(self, id, width, height, REQUEST = None):
        """ Add a display through the web"""
	display = REQUEST.get('id')
        if display is None:
	    raise 'BadRequest', 'The display name does not exist.'
	width = REQUEST.get('width')
        width = int(width)
	if width is None:
	    raise 'BadRequest', 'The display width does not exist.'
	height = REQUEST.get('height')
        height = int(height)
        if display is None:
	    raise 'BadRequest', 'The display height does not exist.'
        self._addDisplay(display, width, height)
        if REQUEST:
            REQUEST['RESPONSE'].redirect(self.absolute_url()+'/manage_displays')
	
    def manage_delDisplay(self, ids, REQUEST = None):
        """ Delete one or more displays through the web"""
        if ids is None:
            return MessageDialog(
                   title='No displays specified',
                   message='No diplsays were selected for deletion!',
                   action ='./manage_displays',)
        for id in ids:
            if not self.displays.has_key(id):
                raise 'BadRequest', (
                      'The display <em>%s</em> does not exist' % id)
	    self._delDisplay(id)

        if REQUEST:
            return self.manage_displays(self, REQUEST)

    def index_html(self, REQUEST, RESPONSE, display = None ):
        """
        The default view of the contents of the Photo

        Returns the contents of the Photo at the requested display size.
	Also, sets the Content-Type HTTP header to the objects content type.
        """
        if display and self.displays.has_key(display):
            # Attempt to handle If-Modified-Since headers.
            ms=REQUEST.get_header('If-Modified-Since', None)
            if ms is not None:
                ms=string.split(ms, ';')[0]
                ms=DateTime(ms).timeTime()
                if self._p_mtime > ms:
                    RESPONSE.setStatus(304)
                    return RESPONSE
            # Set the response header for the photo
            RESPONSE.setHeader('Last-Modified', rfc1123_date(self._p_mtime))
            RESPONSE.setHeader('Content-Type', self.content_type)
	    return getattr(self, 'data_' + display)
        else:
	    # Inherit the Image behavior
            return Image.index_html(self, REQUEST, RESPONSE)
